# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from padacioso import IntentContainer
from ovos_utils.log import LOG
from ovos_utils.messagebus import FakeBus


class PadaciosoContainer:
    def __init__(self, lang: str, cache_dir: str, bus: FakeBus):
        self.cache_dir = cache_dir
        self.lang = lang.lower()
        self.bus = bus
        self.padatious = IntentContainer(False)
        self.bus.on('padatious:register_intent', self.register_intent)
        self.bus.on('padatious:register_entity', self.register_entity)

    def register_intent(self, message):
        """Messagebus handler for registering intents.

        Args:
            message (Message): message triggering action
        """
        lang = message.data.get('lang', self.lang)
        lang = lang.lower()
        if lang == self.lang:
            LOG.debug(f"Loading intent: {message.data['name']}")
            self.padatious.add_intent(message.data['name'],
                                      message.data["samples"])
        else:
            LOG.debug(f"Ignoring {message.data['name']}")

    def register_entity(self, message):
        """Messagebus handler for registering entities.

        Args:
            message (Message): message triggering action
        """
        lang = message.data.get('lang', self.lang)
        lang = lang.lower()
        if lang == self.lang:
            self.padatious.add_entity(message.data['name'],
                                      message.data['samples'])

    def calc_intent(self, utt: str) -> dict:
        intent = self.padatious.calc_intent(utt)
        LOG.debug(intent)
        return intent or dict()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import shutil

from os import environ, getenv
from os.path import dirname, join
from threading import Event
from unittest.mock import Mock
from ovos_utils.messagebus import FakeBus

from neon_minerva.skill import get_skill_object


class SkillTestCase(unittest.TestCase):
    # Define test directories
    test_fs = join(dirname(__file__), "skill_fs")
    data_dir = join(test_fs, "data")
    conf_dir = join(test_fs, "config")
    environ["XDG_DATA_HOME"] = data_dir
    environ["XDG_CONFIG_HOME"] = conf_dir

    # Define static parameters
    bus = FakeBus()
    # Patching FakeBus compat. with MessageBusClient
    bus.emitter = bus.ee
    bus.connected_event = Event()
    bus.connected_event.set()

    bus.run_forever()
    test_skill_id = 'test_skill.test'

    skill = None

    @classmethod
    def setUpClass(cls) -> None:
        # Get test skill
        skill_entrypoint = getenv("TEST_SKILL_ENTRYPOINT")
        if not skill_entrypoint:
            from ovos_plugin_manager.skills import find_skill_plugins
            skill_entrypoints = list(find_skill_plugins().keys())
            assert len(skill_entrypoints) == 1
            skill_entrypoint = skill_entrypoints[0]

        cls.skill = get_skill_object(skill_entrypoint=skill_entrypoint,
                                     skill_id=cls.test_skill_id, bus=cls.bus)
        # Override speak and speak_dialog to test passed arguments
        cls.skill.speak = Mock()
        cls.skill.speak_dialog = Mock()

    def setUp(self):
        self.skill.speak.reset_mock()
        self.skill.speak_dialog.reset_mock()

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(cls.test_fs)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

from os import getenv
from os.path import join, exists
from unittest.mock import Mock

from ovos_bus_client import Message
from ovos_utils.messagebus import FakeBus
from ovos_utils.log import LOG

from neon_minerva.exceptions import IntentException
from neon_minerva.skill import get_skill_object, load_intent_tests
from neon_minerva.intent_services.padatious import PadatiousContainer, TestPadatiousMatcher
from neon_minerva.intent_services.adapt import AdaptContainer
from neon_minerva.intent_services.padacioso import PadaciosoContainer
from neon_minerva.intent_services.common_query import CommonQuery
from neon_minerva.intent_services import IntentMatch


class TestSkillIntentMatching(unittest.TestCase):
    # Static parameters
    bus = FakeBus()
    bus.run_forever()
    test_skill_id = 'test_skill.test'
    padatious_cache = join(getenv("XDG_CACHE_HOME"), "padatious")

    # Define skill and resource spec to use in tests
    valid_intents = load_intent_tests(getenv("INTENT_TEST_FILE"))
    skill_entrypoint = getenv("TEST_SKILL_ENTRYPOINT")

    # Populate configuration
    languages = list(valid_intents.keys())
    core_config_patch = {"secondary_langs": languages}
    negative_intents = valid_intents.pop('unmatched intents', dict())
    common_query = valid_intents.pop("common query", dict())

    # Define intent parsers for tests
    if getenv("TEST_PADACIOSO") == "true":
        container = PadaciosoContainer
    else:
        try:
            from padatious import IntentContainer
            container = PadatiousContainer
        except ImportError as e:
            LOG.error(f"Padatious not installed. Install "
                      f"neon-minerva[padatious] to get Padatious requirements")
            if getenv("TEST_PADACIOSO") is None:
                # Ambiguous request, just use Padacioso
                container = PadaciosoContainer
            else:
                # Explicitly requested Padatious/non-Padacioso
                raise e
    padatious_services = dict()
    adapt_services = dict()
    for lang in languages:
        padatious_services[lang] = container(lang, join(padatious_cache, lang),
                                             bus)
        adapt_services[lang] = AdaptContainer(lang, bus)

    if common_query:
        common_query_service = CommonQuery(bus)

    skill = get_skill_object(skill_entrypoint=skill_entrypoint,
                             skill_id=test_skill_id, bus=bus,
                             config_patch=core_config_patch)

    @classmethod
    def tearDownClass(cls) -> None:
        import shutil
        for service in cls.padatious_services.values():
            try:
                if exists(service.cache_dir):
                    shutil.rmtree(service.cache_dir)
            except Exception as e:
                LOG.exception(e)

    def test_intents(self):
        for lang in self.valid_intents.keys():
            self.assertIsInstance(lang.split('-')[0], str)
            self.assertIsInstance(lang.split('-')[1], str)
            for intent, examples in self.valid_intents[lang].items():
                # TODO: Better method to determine parser?
                if intent.endswith('.intent'):
                    parser = TestPadatiousMatcher(self.padatious_services[lang])
                else:
                    parser = self.adapt_services[lang]

                for utt in examples:
                    if isinstance(utt, dict):
                        data = list(utt.values())[0]
                        utt = list(utt.keys())[0]
                    else:
                        data = list()

                    match = parser.test_intent(utt)
                    self.assertIsInstance(match, IntentMatch)
                    self.assertEqual(match.skill_id, self.test_skill_id)
                    self.assertEqual(match.intent_type,
                                     f"{self.test_skill_id}:{intent}")
                    self.assertEqual(match.utterance, utt)

                    for datum in data:
                        if isinstance(datum, dict):
                            name = list(datum.keys())[0]
                            value = list(datum.values())[0]
                        else:
                            name = datum
                            value = None
                        self.assertIn(name, match.intent_data, utt)
                        if value:
                            self.assertEqual(match.intent_data[name], value)

    def test_negative_intents(self):
        config = self.negative_intents.pop('config', {})
        include_med = config.get('include_med', True)
        include_low = config.get('include_low', False)

        for lang in self.negative_intents.keys():
            adapt = self.adapt_services[lang]
            padatious = TestPadatiousMatcher(self.padatious_services[lang],
                                             include_med=include_med,
                                             include_low=include_low)
            for utt in self.negative_intents[lang]:
                with self.assertRaises(IntentException, msg=utt):
                    adapt.test_intent(utt)
                with self.assertRaises(IntentException, msg=utt):
                    padatious.test_intent(utt)

    def test_common_query(self):
        if not self.common_query:
            return

        qa_callback = Mock()
        qa_response = Mock()
        self.skill.events.add('question:action', qa_callback)
        self.skill.events.add('question:query.response', qa_response)
        for lang in self.common_query.keys():
            for utt in self.common_query[lang]:
                if isinstance(utt, dict):
                    data = list(utt.values())[0]
                    utt = list(utt.keys())[0]
                else:
                    data = dict()
                message = Message('test_utterance',
                                  {"utterances": [utt], "lang": lang})
                self.common_query_service.handle_question(message)
                response = qa_response.call_args[0][0]
                callback = qa_response.call_args[0][0]
                self.assertIsInstance(response, Message)
                self.assertTrue(response.data["phrase"] in utt)
                self.assertEqual(response.data["skill_id"], self.skill.skill_id)
                self.assertIn("callback_data", response.data.keys())
                self.assertIsInstance(response.data["conf"], float)
                self.assertIsInstance(response.data["answer"], str)

                self.assertIsInstance(callback, Message)
                self.assertEqual(callback.data['skill_id'], self.skill.skill_id)
                self.assertEqual(callback.data['phrase'],
                                 response.data['phrase'])
                if not data:
                    continue
                if isinstance(data.get('callback'), dict):
                    self.assertEqual(callback.data['callback_data'],
                                     data['callback'])
                elif isinstance(data.get('callback'), list):
                    self.assertEqual(set(callback.data['callback_data'].keys()),
                                     set(data.get('callback')))
                if data.get('min_confidence'):
                    self.assertGreaterEqual(response.data['conf'],
                                            data['min_confidence'])
                if data.get('max_confidence'):
                    self.assertLessEqual(response.data['conf'],
                                         data['max_confidence'])

    def test_common_play(self):
        # TODO
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import click
import yaml

from os.path import expanduser, relpath, isfile, isdir
from click_default_group import DefaultGroup
from unittest.runner import TextTestRunner
from unittest import makeSuite

from neon_minerva.version import __version__


def _init_tests(debug: bool = False):
    from os.path import join
    from os import makedirs
    from tempfile import mkdtemp
    base_dir = mkdtemp()
    config = join(base_dir, "config")
    data = join(base_dir, "data")
    cache = join(base_dir, "cache")
    makedirs(config, exist_ok=True)
    makedirs(data, exist_ok=True)
    makedirs(cache, exist_ok=True)
    os.environ["XDG_CONFIG_HOME"] = config
    os.environ["XDG_DATA_HOME"] = data
    os.environ["XDG_CACHE_HOME"] = cache

    if debug:
        os.environ["OVOS_DEFAULT_LOG_LEVEL"] = "DEBUG"


def _get_test_file(test_file: str) -> str:
    """
    Parse an input path to locate a test file that may be relative to `~` or the
    current working directory.
    @param test_file: test file argument
    @returns: best guess at the desired file path (may not exist)
    """
    test_file = expanduser(test_file)
    if not isfile(test_file):
        test_file = relpath(test_file)
    return test_file


def _get_skill_entrypoint(skill_entrypoint: str) -> str:
    """
    Parse an input skill entrypoint and resolve either a locally installed skill
    path, or an entrypoint for a plugin skill.
    @param skill_entrypoint: Plugin entrypoint or path to skill
    @returns: absolute file path if exists, else input entrypoint
    """
    skill_path = expanduser(skill_entrypoint)
    if not isdir(skill_path):
        skill_path = relpath(skill_path)
    if isdir(skill_path):
        return skill_path
    return skill_entrypoint


@click.group("minerva", cls=DefaultGroup,
             no_args_is_help=True, invoke_without_command=True,
             help="Minerva: Modular INtelligent Evaluation for a Reliable "
                  "Voice Assistant.\n\n"
                  "See also: minerva COMMAND --help")
@click.option("--version", "-v", is_flag=True, required=False,
              help="Print the current version")
def neon_minerva_cli(version: bool = False):
    if version:
        click.echo(f"Minerva version {__version__}")


@neon_minerva_cli.command
@click.option('--debug', is_flag=True, default=False,
              help="Flag to enable debug logging")
@click.argument("skill_entrypoint")
@click.argument("test_file")
def test_resources(skill_entrypoint, test_file, debug):
    _init_tests(debug)
    os.environ["TEST_SKILL_ENTRYPOINT"] = _get_skill_entrypoint(skill_entrypoint)
    test_file = _get_test_file(test_file)
    if not isfile(test_file):
        click.echo(f"Could not find test file: {test_file}")
        exit(2)
    os.environ["RESOURCE_TEST_FILE"] = test_file
    from neon_minerva.tests.test_skill_resources import TestSkillResources
    TextTestRunner().run(makeSuite(TestSkillResources))


@neon_minerva_cli.command
@click.option('--debug', is_flag=True, default=False,
              help="Flag to enable debug logging")
@click.option('--padacioso', is_flag=True, default=False,
              help="Flag to enable testing with Padacioso instead of Padatious")
@click.argument("skill_entrypoint")
@click.argument("test_file")
def test_intents(skill_entrypoint, test_file, debug, padacioso):
    _init_tests(debug)
    os.environ["TEST_PADACIOSO"] = "true" if padacioso else "false"
    os.environ["TEST_SKILL_ENTRYPOINT"] = _get_skill_entrypoint(skill_entrypoint)
    test_file = _get_test_file(test_file)
    if not isfile(test_file):
        click.echo(f"Could not find test file: {test_file}")
        exit(2)
    os.environ["INTENT_TEST_FILE"] = test_file
    from neon_minerva.tests.test_skill_intents import TestSkillIntentMatching
    TextTestRunner().run(makeSuite(TestSkillIntentMatching))


@neon_minerva_cli.command
@click.option('-l', '--lang', default="en-us",
              help="Language of test_file inputs")
@click.option('-a', '--audio', is_flag=True, default=False,
              help="Test input as audio")
@click.argument("test_file")
def test_utterances(lang, audio, test_file):
    from neon_utils.file_utils import load_commented_file
    from neon_minerva.integration.user_utterance import UtteranceTests

    test_file = _get_test_file(test_file)
    prompts = load_commented_file(test_file).split('\n')
    click.echo(f"Testing {len(prompts)} prompts")
    runner = UtteranceTests(prompts, lang=lang, audio=audio)
    results = runner.run_test()
    click.echo(yaml.safe_dump(results))
